 /*
      This file is part of GNUnet

      GNUnet is free software; you can redistribute it and/or modify
      it under the terms of the GNU General Public License as published
      by the Free Software Foundation; either version 2, or (at your
      option) any later version.

      GNUnet is distributed in the hope that it will be useful, but
      WITHOUT ANY WARRANTY; without even the implied warranty of
      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
      General Public License for more details.

      You should have received a copy of the GNU General Public License
      along with GNUnet; see the file COPYING.  If not, write to the
      Free Software Foundation, Inc., 59 Temple Place - Suite 330,
      Boston, MA 02111-1307, USA.
 */
 
/**
 * @file module/cs.c
 * @brief DHT application protocol using the DHT service.
 *   This is merely for the dht-client library.  The code
 *   of this file is mostly converting from and to TCP messages.
 * @author Marko Rih, Christian Grothoff
 *
 * FIXME: most of this code is still missing.
 */

#include "platform.h"
#include "gnunet_core.h"
#include "gnunet_rpc_service.h"
#include "gnunet_dht_service.h"

/**
 * Global core API.
 */
static CoreAPIForApplication * coreAPI = NULL;

/**
 * Reference to the DHT service API.
 */
static DHT_ServiceAPI * dhtAPI;

/**
 * Information for each table for which persistence is provided
 * by a local client via the TCP link.
 */
typedef struct {
  /**
   * Handle to access the client.
   */
  ClientHandle handler;
  /**
   * For which table is this client responsible?
   */
  DHT_TableId table;
  /**
   * What was the DHT_Datastore that was passed to the DHT service?
   * (must be a pointer since this reference is passed out).
   */
  DHT_Datastore * store;
} CS_TableHandlers;

/**
 * If clients provide a datastore implementation for a table,
 * we keep the corresponding client handler in this array.
 */
static CS_TableHandlers * csHandlers;

/**
 * Size of the csHandlers array.
 */
static unsigned int csHandlersCount;

/**
 * Lock for accessing csHandlers.
 */
static Mutex csLock;

/* ******* implementation of DHT_Datastore via TCP link ********** */

/**
 * Lookup an item in the datastore.
 * @param key the value to lookup
 * @param maxResults maximum number of results
 * @param results where to store the result
 * @return number of results, SYSERR on error
 */
static int tcp_lookup(void * closure,
		      const HashCode160 * key,
		      unsigned int maxResults,
		      DHT_DataContainer * results,
		      int flags) {
  /* send lookup request to matching ClientHandle,
     read reply */
  return SYSERR;
}
  
/**
 * Store an item in the datastore.
 * @param key the key of the item
 * @param value the value to store
 * @return OK if the value could be stored, SYSERR if not (i.e. out of space)
 */
static int tcp_store(void * closure,
		     const HashCode160 * key,
		     const DHT_DataContainer * value,
		     int flags) {
  /* send store request to matching ClientHandle,
     read reply */
  return SYSERR;
}

/**
 * Remove an item from the datastore.
 * @param key the key of the item
 * @param value the value to remove, NULL for all values of the key
 * @return OK if the value could be removed, SYSERR if not (i.e. not present)
 */
static int tcp_remove(void * closure,
		      const HashCode160 * key,
		      const DHT_DataContainer * value,
		      int flags) {
  /* send remove request to matching ClientHandle,
     read reply */
  return SYSERR;
}

/**
 * Iterate over all keys in the local datastore
 *
 * @param processor function to call on each item
 * @param cls argument to processor
 * @return number of results, SYSERR on error
 */
static int tcp_iterate(void * closure,		 
		       int flags,
		       DHT_DataProcessor processor,
		       void * cls) {
  /* send iterate request to matching ClientHandle,
     read replies */
  return SYSERR;
}



/* *********************** CS handlers *********************** */

/**
 * CS handler for joining existing DHT-table.
 */
static int csJoin(ClientHandle client,
                  CS_HEADER * message) {
  CS_TableHandlers * ptr;

  /* FIXME: parse message */
  MUTEX_LOCK(&csLock);
  GROW(csHandlers,
       csHandlersCount,
       csHandlersCount+1);
  ptr = &csHandlers[csHandlersCount-1];
  ptr->store = MALLOC(sizeof(DHT_Datastore));
  ptr->store->iterate = &tcp_iterate;
  ptr->store->remove = &tcp_remove;
  ptr->store->store = &tcp_store;
  ptr->store->lookup = &tcp_lookup;
  ptr->handler = client;
  /* ptr->table = FIXME; */
  MUTEX_UNLOCK(&csLock);
  
  return OK;
}

/**
 * CS handler for leaving DHT-table.
 */
static int csLeave(ClientHandle client,
                   CS_HEADER * message) {
  /* remove client from csHandlers array! */
  return SYSERR;
}

/**
 * CS handler for inserting <key,value>-pair into DHT-table.
 */
static int csPut(ClientHandle client,
		 CS_HEADER * message) {
  /* dhtAPI->store(); */
  return SYSERR;
}

/**
 * CS handler for fetching <key,value>-pairs from DHT-table.
 */
static int csGet(ClientHandle client,
		 CS_HEADER * message) {
  /* dhtAPI->lookup(); */
  return SYSERR;
}

/**
 * CS handler for removing <key,value>-pairs inserted by this node.
 */
static int csRemove(ClientHandle client,
		    CS_HEADER * message) {
  /* dhtAPI->remove(); */
  return SYSERR;
}

/**
 * CS handler for handling exiting client.
 */
static void csClientExit(ClientHandle client) {
  int i;

  MUTEX_LOCK(&csLock);
  for (i=0;i<csHandlersCount;i++) {
    if (csHandlers[i].handler == client) {
      /* csLeave for that table! */
      /* i--; */
    }
  }
  MUTEX_UNLOCK(&csLock);
  LOG(LOG_FAILURE, 
      "%s::csClientExit() - not implemented!\n",
      __FILE__);
}

int initialize_dht_protocol(CoreAPIForApplication * capi) {
  int status;

  dhtAPI = capi->requestService("dht");
  if (dhtAPI == NULL)
    return SYSERR;
  coreAPI = capi;
  LOG(LOG_DEBUG, 
      " DHT registering client handlers: "
      "%d %d %d %d %d\n",
      DHT_CS_PROTO_REQUEST_JOIN,
      DHT_CS_PROTO_REQUEST_LEAVE,
      DHT_CS_PROTO_REQUEST_PUT,
      DHT_CS_PROTO_REQUEST_GET,
      DHT_CS_PROTO_REQUEST_REMOVE);
  status = OK;
  MUTEX_CREATE_RECURSIVE(&csLock);
  if (SYSERR == capi->registerClientHandler(DHT_CS_PROTO_REQUEST_JOIN,
                                            (CSHandler) &csJoin))
    status = SYSERR;
  if (SYSERR == capi->registerClientHandler(DHT_CS_PROTO_REQUEST_LEAVE,
                                            (CSHandler) &csLeave))
    status = SYSERR;
  if (SYSERR == capi->registerClientHandler(DHT_CS_PROTO_REQUEST_PUT,
                                            (CSHandler) &csPut))
    status = SYSERR;
  if (SYSERR == capi->registerClientHandler(DHT_CS_PROTO_REQUEST_GET,
                                            (CSHandler) &csGet))
    status = SYSERR;
  if (SYSERR == capi->registerClientHandler(DHT_CS_PROTO_REQUEST_REMOVE,
                                            (CSHandler) &csRemove))
    status = SYSERR;
  if (SYSERR == capi->registerClientExitHandler(&csClientExit))
    status = SYSERR;
  return status;
}

/**
 * Unregisters handlers, cleans memory structures etc when node exits.
 */
int done_dht_protocol() {
  int status;

  status = OK;
  LOG(LOG_DEBUG, 
      "DHT: shutdown\n");
  if (OK != coreAPI->unregisterClientHandler(DHT_CS_PROTO_REQUEST_JOIN,
					     (CSHandler) &csJoin))
    status = SYSERR;
  if (OK != coreAPI->unregisterClientHandler(DHT_CS_PROTO_REQUEST_LEAVE,
					     (CSHandler) &csLeave))
    status = SYSERR;
  if (OK != coreAPI->unregisterClientHandler(DHT_CS_PROTO_REQUEST_PUT,
					     (CSHandler) &csPut))
    status = SYSERR;
  if (OK != coreAPI->unregisterClientHandler(DHT_CS_PROTO_REQUEST_GET,
					     (CSHandler) &csGet))
    status = SYSERR;
  if (OK != coreAPI->unregisterClientHandler(DHT_CS_PROTO_REQUEST_REMOVE,
					     (CSHandler) &csRemove))
    status = SYSERR;
  if (OK != coreAPI->unregisterClientExitHandler(&csClientExit))
    status = SYSERR;
  /* FIXME: clean up client table! */
  coreAPI->releaseService(dhtAPI);
  dhtAPI = NULL;
  coreAPI = NULL;
  MUTEX_DESTROY(&csLock);
  return status;
}

/* end of cs.c */
